/*
 ___________________________________________________________
/               __           _                              \
|              / _|         (_)                             |
|             | |_ _   _ ___ _  ___  _ __                   |
|             |  _| | | / __| |/ _ \| '_ \                  |
|             | | | |_| \__ \ | (_) | | | |                 |
|             |_|  \__,_|___/_|\___/|_| |_| *               |
|                                                           |
|               The MSX C Library for SDCC                  |
|                   	                                    |
|                                                           |
|                MEGAROM LINKER for FUSION-C                |
|                                                           |
|                  Eric Boez January 2022                   |
|					    	P S G 							|
|                                                           |
\___________________________________________________________/
|
| 	  C source File
|
|	USAGE
|   megalink -c <file name> <size in KB>
|   megalink -c TEST.ROM 48
|        	creation of 48KB empty file named TEST.ROM
|
|   megalink -i <source File> <dest File> <addresse in file in hexdecimal>
|   megalink - DATA.BIN TEST.ROM 8000
|			insert the content of the file DATA.BIN inside TEST.ROM at address 0x8000
|
*/


//#define WINDOWS  /* uncomment this line to compile for windows.*/
# ifdef WINDOWS
# include <direct.h>
# define GETCURRENTDIR _getcwd
# else
# include <unistd.h>
# define GETCURRENTDIR getcwd
# endif

#include<string.h>
# include <stdint.h>
# include <stdio.h>
# include <stdlib.h>
# include <math.h>

# define APP_NAME       " FUSION-C MEGAROM LINKER "
# define VERSION        " 0.1 "
# define DATE           " January 2022 "
# define AUTHOR         " ERIC BOEZ "

///**--------------------------------**
//       True Exit of the Program
//          with error keycodes
//**--------------------------------**
//
void ft_exit( int error, char *text)
{
		    switch (error) {
		        case 0:
		            puts("\nEnding  process...");
		            break;
		            
		        case 1:
		            puts("\nError N°1 Opening file.\n");
		            break;
		            
		        case 2:
		            puts("Error n°2 Loading error.\n");
		            break;
		            
		        case 3:
		            puts("Error n°3 Ouch ! .\n");
		            break;
		            
		        case 4:
		            puts("Wrong argument\n");
		            break;
		            
		        case 5:
		            puts("Error n°5 Init Error\n");
		            break;
		        case 6:
		            puts("Parameters missing");
		            break;
		         case 7:
		            puts("ERROR! :: Cannot allocate memory for the input file ");
		            break;
		        case 8:
		            puts("ERROR! :: Cannot open output file ");
		            break;

		        default:
		            break;
		    }
		    
		    puts(text);
		    puts("\n");

		    exit(EXIT_SUCCESS);
}

///**--------------------------------**
//       Hexadecimal To decimal 
//         conversion
//**--------------------------------**
//
int HexToDec(char *hex) {
		    int p = 0; 
		    int decimal = 0;
		    int r, i;
		    
		    for(i = strlen(hex) - 1 ; i >= 0 ; --i){
		        
		        // converting c[i] to appropriate decimal form
		        if(hex[i]>='0'&&hex[i]<='9'){
		            r = hex[i] - '0';
		        }
		        else{
		            r = hex[i] - 'A' + 10;
		        }
		        
		        decimal = decimal + r * pow(16 , p);
		        ++p;
		    }
		    
		    return decimal;
}
 

///**--------------------------------**
//       Return the lenght 
//           of a file
//**--------------------------------**
//
unsigned int FileLength(char *f)
{
		    FILE *t = fopen(f, "rb");
		    fseek(t, 0, SEEK_END);
		    unsigned int x = ftell(t);
		    fclose(t);
		    return x;
}


///**--------------------------------**
//       Create an empty file
//              on place
//**--------------------------------**
//
void CreateFile(char *filename, unsigned int size)
{
			unsigned int i;
		 	printf("\n Creating file :%s  with a sizeof %d KBYTES",filename,size);
		 	FILE *f = fopen(filename, "wb");

		 	for (i=0; i<size*1024; i++) {
	        		fputc(255, f);
	    	}
	    	fclose(f);
}


///**--------------------------------**
//      Insert content of a file 
//           in another file
//**--------------------------------**
//
void InsertFile(char *source_filename, char *dest_filename, unsigned int address)
{
			unsigned int source_size;
			unsigned int dest_size;

			printf("Inserting file %s in %s at %d",source_filename,dest_filename,address);

			// Get Size of source
			source_size=FileLength(source_filename);
			printf("\n..Source file size : %d",source_size);

			//Temp creation 
			unsigned char *temp = (unsigned char *)malloc(source_size);
			if(temp == NULL)
		    {
		        ft_exit(7,"");
		    }

		    // Read the entire Source file
		    FILE *in = fopen(source_filename, "rb");
		    fread(temp, source_size, 1, in);
		    fclose(in);



		    //  temp creation
		    dest_size=FileLength(dest_filename);
		    unsigned char *temp_dest = (unsigned char *)malloc(dest_size);
			if(temp_dest == NULL)
		    {
		        ft_exit(7,"");
		    }
		    FILE *in2 = fopen(dest_filename, "rb");
		    fread(temp_dest, dest_size, 1, in);
		    fclose(in2);

		    // Insert Data in dest
		    for (int i = 0; i < source_size; ++i)
		    {

		    	temp_dest[address+i]=temp[i];
		    }

		   
		   	//write to Destination
		    FILE *out = fopen(dest_filename, "wb");

		    if(out == NULL)
    		{
        			 ft_exit(8,"");
    		}
    		

    		fwrite(temp_dest, dest_size, 1, out);
    		fclose(out);

		    free(temp);
		    free(temp_dest);
}

///**--------------------------------**
//             MAIN 
//          
//**--------------------------------**
//
int main(int argc, char *argv[]) {
	

	unsigned int size;
	unsigned int address;


	printf("\n------------------------------------------------------\n");
	printf(APP_NAME);
	printf(VERSION);
	printf("\n------------------------------------------------------\n");

	if (argc<4)
    {
     
        ft_exit(6,"");
    }

		switch (argv[1][1])
		{
			        case 'c':								// create File
			         	size=atoi(argv[3]);
			             
			           	CreateFile(argv[2],size);
			            break;
			                
			        case 'i':								// insert binarie
			           	if (argc<5)
   						{
        					ft_exit(6,"");
    					}
			           	//address=atoi(argv[4]);				
			           	address=HexToDec(argv[4]);
			           	InsertFile(argv[2],argv[3],address);
			            break;
			                
			          
			         default:       
			        		ft_exit(4,"");
		}


		exit(EXIT_SUCCESS);
}